{-# LANGUAGE DerivingStrategies #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE TypeApplications #-}
{-|
Module      : Gargantext.Core.LinearAlgebra
Description : Linear Algebra utility functions
Copyright   : (c) CNRS, 2017-Present
License     : AGPL + CECILL v3
Maintainer  : team@gargantext.org
Stability   : experimental
Portability : POSIX

Linear algebra utility functions to be used across all the Gargantext modules requiring it.
-}

module Gargantext.Core.LinearAlgebra (
  -- * Types
    Index(..)

  -- * Functions
  , createIndices

  -- * Convertion functions
  , accelerate2MassivMatrix
  , massiv2AccelerateMatrix
  , massiv2AccelerateVector

  -- * Operations on matrixes
  , (.*)
  , diag
  , termDivNan
  , distributional
  ) where

import Data.Array.Accelerate qualified as Acc
import Data.Bimap (Bimap)
import Data.Bimap qualified as Bimap
import Data.List.Split qualified as Split
import Data.Map.Strict (Map)
import Data.Map.Strict qualified as M
import Data.Massiv.Array (D, Matrix, Vector, Array, Ix3)
import Data.Massiv.Array qualified as A
import Data.Set qualified as S
import Data.Set (Set)
import Prelude
import Protolude.Safe (headMay)

newtype Index = Index { _Index :: Int }
  deriving newtype (Eq, Show, Ord, Num, Enum)

createIndices :: Ord t => Map (t, t) b -> Bimap Index t
createIndices = set2indices . map2set
  where
    map2set :: Ord t => Map (t, t) a -> Set t
    map2set cs' = foldr (\(t1, t2) s -> S.insert t1 $! S.insert t2 $! s) mempty $ M.keys cs'

    set2indices :: Ord t => Set t -> Bimap Index t
    set2indices s = foldr (uncurry Bimap.insert) Bimap.empty (zip [0..] $ S.toList s)

-- | Converts an accelerate matrix into a Massiv matrix.
accelerate2MassivMatrix :: (A.Unbox a, Acc.Elt a) => Acc.Matrix a -> Matrix D a
accelerate2MassivMatrix m =
  let (Acc.Z Acc.:. _r Acc.:. c) = Acc.arrayShape m
  in A.delay $ A.fromLists' @A.U A.Par $ Split.chunksOf c (Acc.toList m)

-- | Converts a massiv matrix into an accelerate matrix.
massiv2AccelerateMatrix :: (Acc.Elt a, A.Source r a) => Matrix r a -> Acc.Matrix a
massiv2AccelerateMatrix m =
  let m' = A.toLists2 m
      r  = Prelude.length m'
      c  = maybe 0 Prelude.length (headMay m')
  in Acc.fromList (Acc.Z Acc.:. r Acc.:. c) (mconcat m')

-- | Converts a massiv vector into an accelerate one.
massiv2AccelerateVector :: (A.Source r a, Acc.Elt a) => A.Vector r a -> Acc.Vector a
massiv2AccelerateVector m =
  let m' = A.toList m
      r  = Prelude.length m'
  in Acc.fromList (Acc.Z Acc.:. r) m'


-- | Computes the diagnonal matrix of the input one.
diag :: Num e => Matrix D e -> Vector D e
diag matrix =
  let (A.Sz2 rows _cols) = A.size matrix
      newSize = A.Sz1 rows
  in A.backpermute' newSize (\j -> j A.:. j) matrix



-- | `distributional m` returns the distributional distance between terms each
-- pair of terms as a matrix.  The argument m is the matrix $[n_{ij}]_{i,j}$
-- where $n_{ij}$ is the coocccurrence between term $i$ and term $j$.
--
-- ## Basic example with Matrix of size 3: 
--
-- >>> theMatrixInt 3
-- Matrix (Z :. 3 :. 3)
--   [ 7, 4, 0,
--     4, 5, 3,
--     0, 3, 4]
--
-- >>> distributional $ theMatrixInt 3
-- Matrix (Z :. 3 :. 3)
--   [ 1.0, 0.0, 0.9843749999999999,
--     0.0, 1.0,                0.0,
--     1.0, 0.0,                1.0]
--
-- ## Basic example with Matrix of size 4: 
--
-- >>> theMatrixInt 4
-- Matrix (Z :. 4 :. 4)
--   [ 4, 1, 2, 1,
--     1, 4, 0, 0,
--     2, 0, 3, 3,
--     1, 0, 3, 3]
--
-- >>> distributional $ theMatrixInt 4
-- Matrix (Z :. 4 :. 4)
--   [                  1.0,                   0.0, 0.5714285714285715, 0.8421052631578947,
--                      0.0,                   1.0,                1.0,                1.0,
--     8.333333333333333e-2,             4.6875e-2,                1.0,               0.25,
--       0.3333333333333333, 5.7692307692307696e-2,                1.0,                1.0]
--
-- /IMPORTANT/: As this function computes the diagonal matrix in order to carry on the computation
-- the input has to be a square matrix, or this function will fail at runtime.
distributional :: forall e. (Ord e, Fractional e, Num e)
               => Matrix D Int
               -> Matrix D e
distributional m' = result
 where
    m :: Matrix D e
    m = A.map fromIntegral m'
    n = dim m'

    diag_m :: Vector D e
    diag_m = diag m

    -- replicate (constant (Z :. n :. All)) diag_m
    d_1 :: Matrix D e
    d_1 = let (A.Sz1 r) = A.size diag_m
          in A.backpermute' (A.Sz2 n r) (\(_ A.:. i) -> i) diag_m

    -- replicate (constant (Z :. All :. n)) diag_m
    d_2 :: Matrix D e
    d_2 = let (A.Sz1 r) = A.size diag_m
          in A.backpermute' (A.Sz2 r n) (\(i A.:. _) -> i) diag_m

    mi :: Matrix D e
    mi = (.*) (termDivNan m d_1) (termDivNan m d_2)

    -- The matrix permutations is taken care of below by directly replicating
    -- the matrix mi, making the matrix w unneccessary and saving one step.
    -- replicate (constant (Z :. All :. n :. All)) mi
    w_1 :: Array D Ix3 e
    w_1 = let (A.Sz2 r c) = A.size mi
          in A.backpermute' (A.Sz3 r n c) (\(x A.:> _y A.:. z) -> x A.:. z) mi

    -- replicate (constant (Z :. n :. All :. All)) mi
    w_2 :: Array D Ix3 e
    w_2 = let (A.Sz2 r c) = A.size mi
          in A.backpermute' (A.Sz3 n r c) (\(_x A.:> y A.:. z) -> y A.:. z) mi

    w' :: Array D Ix3 e
    w' = A.zipWith min w_1 w_2

    -- The matrix ii = [r_{i,j,k}]_{i,j,k} has r_(i,j,k) = 0 if k = i OR k = j 
    -- and r_(i,j,k) = 1 otherwise (i.e. k /= i AND k /= j).
    -- generate (constant (Z :. n :. n :. n)) (lift1 (\( i A.:. j A.:. k) -> cond ((&&) ((/=) k i) ((/=) k j)) 1 0))
    ii :: Array D Ix3 e
    ii = A.makeArrayR A.D A.Par (A.Sz3 n n n) $ \(i A.:> j A.:. k) -> if k /= i && k /= j then 1 else 0

    z_1 :: Matrix D e
    z_1 = sumRows ((.*) w' ii)

    z_2 :: Matrix D e
    z_2 = sumRows ((.*) w_1 ii)

    result = termDivNan z_1 z_2

-- | Term by term division where divisions by 0 produce 0 rather than NaN.
termDivNan :: (Eq a, Fractional a) => Matrix D a -> Matrix D a -> Matrix D a
termDivNan = A.zipWith (\i j -> if j == 0 then 0 else i / j)

sumRows :: Num e => Array D A.Ix3 e -> Array D A.Ix2 e
sumRows matrix =
  let A.Sz3 rows cols z = A.size matrix
  in A.makeArray (A.getComp matrix) (A.Sz2 rows cols) $ \(i A.:. j) ->
       A.sum (A.backpermute' (A.Sz1 z) (\c -> i A.:> j A.:. c) matrix)

-- | Matrix cell by cell multiplication
(.*) :: (A.Index ix, Num a)
     => Array D ix a
     -> Array D ix a
     -> Array D ix a
(.*) = A.zipWith (*)

-- | Get the dimensions of a /square/ matrix.
dim :: A.Size r => Matrix r a -> Int
dim m = n
  where
    (A.Sz2 _ n) = A.size m
