{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE MultiWayIf #-}
{-# LANGUAGE DerivingStrategies #-}
{-# LANGUAGE ScopedTypeVariables #-}

module Test.Core.LinearAlgebra where

import Data.Array.Accelerate hiding (Ord, Eq, map, (<=))
import Data.Array.Accelerate.Interpreter qualified as Naive
import Data.Bifunctor (first)
import Data.Bimap (Bimap)
import Data.Bimap qualified as Bimap
import Data.Map.Strict (Map)
import Data.Map.Strict qualified as M
import Data.Massiv.Array qualified as Massiv
import Gargantext.Core.LinearAlgebra qualified as LA
import Gargantext.Core.Methods.Matrix.Accelerate.Utils qualified as Legacy
import Gargantext.Core.Methods.Similarities.Accelerate.Distributional qualified as Legacy
import Gargantext.Core.Viz.Graph.Index qualified as Legacy
import Gargantext.Orphans.Accelerate (sliceArray)
import Gargantext.Prelude (headMay)
import Prelude hiding ((^))
import qualified Data.Array.Accelerate as A
import qualified Data.List.Split.Internals as Split
import Test.Tasty
import Test.Tasty.QuickCheck
import Data.Proxy
import Data.Scientific


--
-- Utility types and functions
--

newtype SquareMatrix a = SquareMatrix { _SquareMatrix :: Matrix a }
  deriving newtype (Show, Eq)

instance (Elt a, Show a, Prelude.Num a, Ord a, Arbitrary a) => Arbitrary (SquareMatrix a) where
  arbitrary = do
    x <- choose (1,30)
    let sh = Z :. x :. x
    SquareMatrix . A.fromList sh <$> vectorOf (x*x) arbitrary
  shrink = map (SquareMatrix) . sliceArray . _SquareMatrix

compareImplementations :: (Arbitrary a, Eq b, Show b)
                       => (a -> b)
                       -> (a -> c)
                       -> (c -> b)
                       -> a
                       -> Property
compareImplementations implementation1 implementation2 mapResults inputData
  = implementation1 inputData === mapResults (implementation2 inputData)

compareImplementations' :: (Arbitrary a, Eq c, Show c)
                       => (a -> b)
                       -> (a -> b)
                       -> (b -> c)
                       -> a
                       -> Property
compareImplementations' implementation1 implementation2 mapResults inputData
  = mapResults (implementation1 inputData) === mapResults (implementation2 inputData)

mapCreateIndices :: Ord t => (Map t Legacy.Index, Map Legacy.Index t) -> Bimap LA.Index t
mapCreateIndices (_m1, m2) = Bimap.fromList $ map (first LA.Index) $ M.toList m2

type TermDivNanShape = Z :. Int :. Int

twoByTwo :: SquareMatrix Int
twoByTwo = SquareMatrix $ fromList (Z :. 2 :. 2) (Prelude.replicate 4 0)

testMatrix_01 :: SquareMatrix Int
testMatrix_01 = SquareMatrix $ fromList (Z :. 14 :. 14) $
  [  30,  36, -36, -16,   0,   7,  34,  -7,   5,  -4,   0,  21,   6, -35,
      0, -31,  20, -15, -22,  -7, -22, -37, -29, -29,  23, -31, -29, -23,
    -24, -29,  19,  -6,  16,   7,  15, -27, -27, -30,  -9, -33,  18, -23,
      7, -36,  12,  26, -17,  -3,  -2, -15,  -4,  26,  24,   9,  -4,   4,
     32,  28,  -2, -10,  34,  -3,  20,  -9, -22,  20, -26,  34,  18, -21,
      7, -12,  12,  -2,  36,  10,  34, -37,  13,  -9, -28,  34,  33, -18,
     -4, -32,  -1,  29,  29, -28,  24,  28,  35,  19,   8, -18,  25, -35,
    -14,  -4, -24,  -1,   7,  34, -37, -28, -12, -32,  -5, -23,  27,  33,
    -36, -28,  21, -29,  -2, -26,  -4, -31, -26, -21,  33, -11, -33,  20,
     25,  14,   5,  -7,   5,  24,  37,   1,  -3,  23,  25, -16,  17,   5,
    -35,  36,  -2,  -2,   1, -14,  34, -30, -10,  12,  25,  21,   0,  34,
     17,  -1,  20, -19,  15,  20,  -5, -30, -35, -13,   5,  17, -10, -19,
    -34, -11, -18,  26, -29, -28,   0,   3,  23,  -6,  36,   4,  16,  28,
     13, -37, -16,   2,   7, -13,  21, -10, -33, -33, -26, -19,  -1,  29]

testMatrix_02 :: SquareMatrix Int
testMatrix_02 = SquareMatrix $ fromList (Z :. 7 :. 7) $
  [  30,  36, -36, -16,   0,   7,  34,
      0, -31,  20, -15, -22,  -7, -22,
    -24, -29,  19,  -6,  16,   7,  15,
      7, -36,  12,  26, -17,  -3,  -2,
     32,  28,  -2, -10,  34,  -3,  20,
      7, -12,  12,  -2,  36,  10,  34,
     13, -37, -16,   2,   7, -13,  21]


--
-- Main test runner
--

-- | Needed as the LLVM and Naive backend generates some double with a long exponent which
-- won't compare verbatim.
tests :: TestTree
tests = testGroup "LinearAlgebra" [
      testProperty "createIndices roundtrip" (compareImplementations (LA.createIndices @Int @Int) Legacy.createIndices mapCreateIndices)
    , testProperty "termDivNan"     compareTermDivNan
    , testProperty "diag"           compareDiag
    , testGroup "distributional" [
        testProperty "2x2" (compareDistributional (Proxy @Double) twoByTwo)
      , testProperty "7x7" (compareDistributional (Proxy @Double) testMatrix_02)
      , testProperty "14x14" (compareDistributional (Proxy @Double) testMatrix_01)
      , testProperty "roundtrips" (compareDistributional (Proxy @Double))
      ]
    ]

--
-- Tests
--

compareTermDivNan :: (Array TermDivNanShape Double)
                  -> (Array TermDivNanShape Double)
                  -> Property
compareTermDivNan i1 i2
  = let massiv     = LA.termDivNan (accelerate2MassivMatrix i1) (accelerate2MassivMatrix i2)
        accelerate = Naive.run (Legacy.termDivNan (use i1) (use i2))
    in accelerate === massiv2AccelerateMatrix massiv

compareDiag :: SquareMatrix Int -> Property
compareDiag (SquareMatrix i1)
  = let massiv     = Massiv.computeAs Massiv.U $ LA.diag (accelerate2MassivMatrix i1)
        accelerate =  Naive.run (Legacy.diag (use i1))
    in accelerate === massiv2AccelerateVector massiv

compareDistributional :: forall e.
                      ( Eq e
                      , Show e
                      , FromIntegral Int e
                      , Prelude.RealFloat e
                      , A.Ord e
                      , Ord e
                      , Prelude.Fractional (Exp e)
                      , Prelude.Fractional e
                      ) => Proxy e
                      -> SquareMatrix Int
                      -> Property
compareDistributional Proxy (SquareMatrix i1)
  = let massiv     = Massiv.computeAs Massiv.B $ LA.distributional @e (accelerate2MassivMatrix i1)
        accelerate = Legacy.distributionalWith Naive.run i1
        expected   = map conv (A.toList accelerate)
        actual     = map conv (mconcat (Massiv.toLists2 massiv))
    in counterexample "size not equal" (Prelude.length expected === Prelude.length actual) .&&. expected === actual
  where
    conv :: e -> Scientific
    conv = fromFloatDigits
